/**
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package
{
	import flash.display.*
	import flash.events.*
	import flash.media.Sound;
	import flash.media.SoundChannel;
	import flash.net.*

	[SWF(width=800, height=600, frameRate=255, backgroundColor=0x000000)]
	/**
	 * An example on using JITB for GPU-accelerated Pixel Bender filters: raytracing a Julia Set.
	 * The code may look a bit peculiar in places, due to JITB development being in progress at the time
	 * of writing.
	 *
	 * WARNING: running this in a normal Flash Player will likely cause it to crash (or take 2 minutes to render
	 * the first frame ;) ). Also, if your GPU is famed for overheating, please avoid running this.
	 *
	 * JITB by Joa Ebert, instructions: http://blog.joa-ebert.com/2010/10/06/apparat-1-0-rc7-is-here/
	 *
	 * @author David Lenaerts
	 *
	 */
	public class JITBJuliaRaytracer extends Sprite
	{
		// there's a few pixels in the corner like this... toolkit on gpu does it okay on 800x600
		private var _width : Number;
		private var _height : Number;

		private var t : Number;

		private var _juliaContainer : Sprite;
		private var _sound : Sound;
		private var _channel : SoundChannel;
		private var _loader : URLLoader;
		private var _shader : Shader;
		// properties of the music
		private var _peak : Number;
		private var _peak2 : Number;
		// properties of the set:
		private var _w : Number;
		private var _rotY : Number;

		public function JITBJuliaRaytracer()
		{
			init();
		}

		private function init() : void
		{
			_width = 800.0;
			_height = 600.0;
			_loader = new URLLoader();
			_loader.dataFormat = URLLoaderDataFormat.BINARY;
			_loader.addEventListener(Event.COMPLETE, onLoadComplete);
			_loader.load(new URLRequest("RTJulia.pbj"));
			t = 0.0;
		}


		public function onLoadComplete(event : Event) : void
		{
			_rotY = 0;
			_juliaContainer = new Sprite();
			addChild(_juliaContainer);
			var c: Array = new Array();
			c[0] = -0.4;
			c[1] = 0.0;//-0.68;
			c[2] = -0.36;
			c[3] = -0.04;
			_shader = new Shader();
			_shader.byteCode = _loader.data;
			_shader.data.width.value = [ _width ];
			_shader.data.height.value = [ _height ];
			_shader.data.c.value = c;
			_shader.data.camDistance.value = [ 3 ];
			_shader.data.rotationY.value = [ 0 ];
			_w = -1.0;
			_shader.data._w.value = [ 0 ];

			_juliaContainer.graphics.clear();
			_juliaContainer.graphics.beginShaderFill(_shader);
			_juliaContainer.graphics.drawRect(0.0, 0.0, _width, _height);
			_juliaContainer.graphics.endFill();

			_peak = 0;

			_sound = new Sound(new URLRequest("Xtal.mp3"));
			_channel = _sound.play();

			addEventListener(Event.ENTER_FRAME, onEnterFrame);
		}

		public function onEnterFrame(event : Event) : void
		{
			t += 0.01;
			
			var c: Array = new Array();

			if (_channel.rightPeak < .7) {
				_peak = _peak*.8 + _channel.rightPeak*.2*.2;
			}
			else {
				_peak = _peak*.8 + _channel.rightPeak*.4*.2;
			}
			
			if (_channel.leftPeak < .7) {
				_peak2 = _peak2*.99;
			}
			else {
				_peak2 = _peak2*.7 + _channel.leftPeak*.4*.3;
			}

			c[0] = Math.cos(t*.55)*-.4 + _peak2;
			c[1] = Math.cos(t*.66)*.5;
			c[2] = _peak;//Math.cos(t*.77)*.15 - 0.36;
			c[3] = Math.sin(t*.3);

			_w = _w + .0005;
			_w = Math.min(_w, 0.0);

			_rotY += .004;
			if (_channel.leftPeak > .8) _rotY -= .01;
			_shader.data.c.value = c;
			_shader.data.rotationY.value = [ _rotY ];
			_shader.data._w.value = [ _w ];
		}
	}
}