/**
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package
{
	import flash.display.Bitmap;
	import flash.display.Loader;
	import flash.display.Shader;
	import flash.display.Sprite;
	import flash.events.Event;
	import flash.net.URLLoader;
	import flash.net.URLLoaderDataFormat;
	import flash.net.URLRequest;

	[SWF(width="800", height="450", frameRate="120")]
	/**
	 * An example on using JITB for GPU-accelerated Pixel Bender filters: raymarching a heightmap terrain.
	 * The code may look a bit peculiar in places, due to JITB development being in progress at the time
	 * of writing.
	 *
	 * WARNING: running this in a normal Flash Player will likely cause it to crash (or take 2 minutes to render
	 * the first frame ;) ). Also, if your GPU is famed for overheating, please avoid running this.
	 *
	 * JITB by Joa Ebert, instructions: http://blog.joa-ebert.com/2010/10/06/apparat-1-0-rc7-is-here/
	 *
	 * @author David Lenaerts
	 *
	 */
	public class TerrainMarcher extends Sprite
	{
		private var _width : Number;
		private var _height : Number;

		private var _imageURLs : Array;
		private var _loadIndex : int;
		private var _loader : Loader;
		private var _bitmapDatas : Array;
		private var _urlLoader : URLLoader;
		private var _shader : Shader;
		private var _t : Number;
		private var _camX : Number;
		private var _camY : Number;
		private var _camZ : Number;
		private var _diffuseR : Number;
		private var _diffuseG : Number;
		private var _diffuseB : Number;
		private var _specularR : Number;
		private var _specularG : Number;
		private var _specularB : Number;
		private var _ambientR : Number;
		private var _ambientG : Number;
		private var _ambientB : Number;
		private var _lightX : Number;
		private var _lightY : Number;
		private var _lightZ : Number;
		private var _rotationX : Number;
		private var _rotationY : Number;

		public function TerrainMarcher()
		{
			_width = 800.0;
			_height = 450.0;
			_camX = 110.0;
			_camY = 200.0;
			_camZ = -300.0;
			_diffuseR = 1.0;
			_diffuseG = 0.56;
			_diffuseB = 0.2;
			_specularR = .78;
			_specularG = .9;
			_specularB = .0;
			_ambientR = 0.02;
			_ambientG = 0.02;
			_ambientB = 0.05;
			_lightX = -1.0;
			_lightY = 1.0;
			_lightZ = 1.0;
			_rotationX = 0.3;
			_rotationY = 0.0;

			loadImages();
		}

		public function loadImages() : void
		{
			_imageURLs = new Array();
			_bitmapDatas = new Array();
			_imageURLs[0] = "bigmapBlur.jpg";
			_imageURLs[1] = "diffuseMapBig.jpg";
			_imageURLs[2] = "bigmap_NRM.png";
			_imageURLs[3] = "sky_sphere4.jpg";
			_loadIndex = 0;
			loadNext();
		}

		public function loadNext() : void
		{
			var url : String = _imageURLs[_loadIndex];
			var urlRequest : URLRequest = new URLRequest(url);
			_loader = new Loader();
			_loader.contentLoaderInfo.addEventListener(Event.COMPLETE, onImageLoadComplete);
			_loader.load(urlRequest);
		}

		public function onImageLoadComplete(event : Event) : void
		{
			var bitmap : Bitmap = Bitmap(_loader.content);

			_bitmapDatas[_loadIndex] = bitmap.bitmapData;
			_loadIndex = _loadIndex + 1;
			if(_loadIndex == 4) {
				loadShader();
			}
			else {
				loadNext();
			}
		}

		public function loadShader() : void
		{
			_urlLoader = new URLLoader();
			_urlLoader.dataFormat = URLLoaderDataFormat.BINARY;
			_urlLoader.addEventListener(Event.COMPLETE, onShaderLoadComplete);
			_urlLoader.load(new URLRequest("RTTerrain.pbj"));
		}

		public function onShaderLoadComplete(event : Event) : void
		{
			var arr : Array;
			_shader = new Shader(_urlLoader.data);
			arr = new Array();
			arr[0] = _width;
			_shader.data.width.value = arr;
			arr = new Array();
			arr[0] = _height;
			_shader.data.height.value = arr;
			_shader.data.heightMap.input = _bitmapDatas[0];
			_shader.data.diffuseMap.input = _bitmapDatas[1];
			_shader.data.normalMap.input = _bitmapDatas[2];
			_shader.data.sphereMap.input = _bitmapDatas[3];

			arr = new Array();
			arr[0] = 900.0;
			_shader.data.fogDist.value = arr;
			arr = new Array();
			arr[0] = 600.0;
			_shader.data.minFogDist.value = arr;
			arr = new Array();
			arr[0] = 130.0;
			_shader.data.elevation.value = arr;
			arr = new Array();
			arr[0] = 1.0;
			arr[1] = 1.0;
			_shader.data.mapScale.value = arr;
			arr = new Array();
			arr[0] = _camX;
			arr[1] = _camY;
			arr[2] = _camZ;
			_shader.data.camPosition.value = arr;
			arr = new Array();
			arr[0] = _rotationX;
			_shader.data.rotationX.value = arr;
			arr = new Array();
			arr[0] = _rotationY;
			_shader.data.rotationY.value = arr;
			arr = new Array();
			arr[0] = _lightX;
			arr[1] = _lightY;
			arr[2] = _lightZ;
			_shader.data.light.value = arr;
			arr = new Array();
			arr[0] = _diffuseR;
			arr[1] = _diffuseG;
			arr[2] = _diffuseB;
			arr[3] = 0.0;
			_shader.data.diffuseColor.value = arr;
			arr = new Array();
			arr[0] = _specularR;
			arr[1] = _specularG;
			arr[2] = _specularB;
			arr[3] = 0.0;
			_shader.data.specularColor.value = arr;
			arr = new Array();
			arr[0] = _ambientR;
			arr[1] = _ambientG;
			arr[2] = _ambientB;
			arr[3] = 0.0;
			_shader.data.ambient.value = arr;

			_t = 0.0;
			graphics.beginShaderFill(_shader);
			graphics.drawRect(0.0, 0.0, _width, _height);
			graphics.endFill();
			addEventListener(Event.ENTER_FRAME, onEnterFrame);
		}

		public function onEnterFrame(event : Event) : void
		{
			var arr : Array;
			_t += .007;
			_camX = Math.sin(_t)*300.0 - 300;
			_rotationY = 0.0-Math.sin(_t)*.9;
			_camY -= .7;
			_rotationX -= .002;
			_camY = Math.max(_camY, 110);
			_rotationX = Math.max(_rotationX, -.3);
			_camZ += 2.0;
			arr = new Array();
			arr[0] = _rotationX;
			_shader.data.rotationX.value = arr;
			arr = new Array();
			arr[0] = _rotationY;
			_shader.data.rotationY.value = arr;

			arr = new Array();
			arr[0] = _camX;
			arr[1] = _camY;
			arr[2] = _camZ;
			_shader.data.camPosition.value = arr;

			if (_camY == 110) {
				updateLights();
			}
		}

		private function updateLights() : void
		{
			var arr : Array = new Array();
			_lightY = _lightY - .005;
			_lightY = Math.max(-.7, _lightY);

			arr[0] = _lightX;
			arr[1] = _lightY;
			arr[2] = _lightZ;
			_shader.data.light.value = arr;


			_diffuseG += (.8 - .56)*.007;
			_diffuseB += (.5 - .2)*.007;
			_specularR += (1.0 - .78)*.007;
			_specularG += (1.0 - .9)*.007;
			_specularB += (1.0 - .0)*.007;
			_ambientR += (.05 - .02)*.007;
			_ambientG += (.05 - .02)*.007;
			_ambientB += (.1 - .05)*.007;

			_diffuseG = Math.min(.8, _diffuseG);
			_diffuseB = Math.min(.5, _diffuseB);
			_specularR = Math.min(1.0, _specularR);
			_specularG = Math.min(1.0, _specularG);
			_specularB = Math.min(1.0, _specularB);
			_ambientR = Math.min(0.05, _ambientR);
			_ambientG = Math.min(0.05, _ambientG);
			_ambientB = Math.min(0.1, _ambientB);

			arr = new Array();
			arr[0] = _diffuseR;
			arr[1] = _diffuseG;
			arr[2] = _diffuseB;
			arr[3] = 0.0;
			_shader.data.diffuseColor.value = arr;
			arr = new Array();
			arr[0] = _specularR;
			arr[1] = _specularG;
			arr[2] = _specularB;
			arr[3] = 0.0;
			_shader.data.specularColor.value = arr;
			arr = new Array();
			arr[0] = _ambientR;
			arr[1] = _ambientG;
			arr[2] = _ambientB;
			arr[3] = 0.0;
			_shader.data.ambient.value = arr;
		}
	}
}
